/**
@file dkcMTRand32.c
@brief Mersenne Twister 19937 implementation
*/
#define DKUTIL_C_MT_RAND32_C
#include "dkcMTRand32.h"
#include "cpu_x86.h"
#include <dkutil_c/dkc_misc.h>

/////////////////////////////////////////////////////////////////////////
//in mymt.c

/* Period parameters */  
//#define dkcdMT_RAND32_N 624 in header
#define M 397
#define MATRIX_A 0x9908B0DF   /* constant vector a */
#define UMASK 0x80000000 /* most significant w-r bits */
#define LMASK 0x7FFFFFFF /* least significant r bits */
#define MIXBITS(u,v) ( ((u) & UMASK) | ((v) & LMASK) )
#define TWIST(u,v) ((MIXBITS(u,v) >> 1) ^ ((v)&1 ? MATRIX_A : 0))


/////////////////////////////////////////////////////////////////////////

/* defined in mymt_x86.nas */
/*
DKC_EXTERN void myMT_Next_SSE2(DKC_MT_RAND32 *mt);
DKC_EXTERN void myMT_Next_MMX(DKC_MT_RAND32 *mt);
*/
//void myMT_Next_other(myMT_t *mt);

void dkcMTRand32Next_C(DKC_MT_RAND32 *mt)
{
	int i;
	uint32 *p, *q;

	p = mt->state;
	for (i = 0; i < dkcdMT_RAND32_N - M; i++, p++) {
		*p = p[M] ^ TWIST(p[0], p[1]);
	}

	for (i = 0; i < M - 1; i++, p++) {
		*p = p[M-dkcdMT_RAND32_N] ^ TWIST(p[0], p[1]);
	}

	*p = p[M-dkcdMT_RAND32_N] ^ TWIST(p[0], mt->state[0]);
	mt->index = 0;

	p = mt->state;
	q = mt->output;
	for (i = 0; i < dkcdMT_RAND32_N; i++) {
		uint32 y;
		y = *p++;
		/* Tempering */
		y ^= (y >> 11);
		y ^= (y << 7) & 0x9D2C5680;
		y ^= (y << 15) & 0xEFC60000;
		y ^= (y >> 18);
		*q++ = y;
	}
}

static DKC_INLINE next(DKC_MT_RAND32 *mt)
{
	DKC_MT_RAND32_GENERATE_F_TYPE pf;
	pf = mt->function_ptr;
	pf(mt);
}
		
//myMT_SetSeed()
void WINAPI dkcMTRand32Init(DKC_MT_RAND32 *mt, uint32 seed)
{
	uint32 *p;
	int i;

	p = mt->state;
	p[0] = seed;

	for (i = 1; i < dkcdMT_RAND32_N; i++) {
		p[i] = (1812433253 * (p[i-1] ^ (p[i-1] >> 30)) + i); 
		/* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
		/* In the previous versions, MSBs of the seed affect   */
		/* only MSBs of the array state[]. */
		/* 2002/01/09 modified by Makoto Matsumoto */
		p[i] &= 0xffffffffUL;
    /* for >32 bit machines */
	}
	//update
	mt->index = i;
	//myMT_Next(mt);
}

void WINAPI dkcMTRand32InitByArray(DKC_MT_RAND32 *p,uint32 *init_key,int32 key_length)
{
  int i, j, k;
	uint32 *mt = p->state;
	//uint32 mti = p->index;
  dkcMTRand32Init(p,19650218UL);
  i=1; j=0;
  k = (dkcdMT_RAND32_N>key_length ? dkcdMT_RAND32_N : key_length);
  for (; k; k--) {
      mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1664525UL))
        + init_key[j] + j; /* non linear */
      mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
      i++; j++;
      if (i>=dkcdMT_RAND32_N) { mt[0] = mt[dkcdMT_RAND32_N-1]; i=1; }
      if (j>=key_length) j=0;
  }
  for (k=dkcdMT_RAND32_N-1; k; k--) {
      mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1566083941UL))
        - i; /* non linear */
      mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
      i++;
      if (i>=dkcdMT_RAND32_N) { mt[0] = mt[dkcdMT_RAND32_N-1]; i=1; }
  }

  mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */ 
	//update
	//next(p);
	//p->index = mti;
}


#if 0
void myMT_SetSeeds(myMT_t *mt, uint32 *seed, int seedNum)
{
	uint32 *p = mt->state;
	int i;
	int num = (seedNum < dkcdMT_RAND32_N) ? seedNum : dkcdMT_RAND32_N;

	if (num <= 0) {
		myMT_SetSeed(mt, 1234);
		return;
	}

	memcpy(p, seed, num * sizeof(uint32));

	for (i = num; i < dkcdMT_RAND32_N; i++) {
		p[i] = (1812433253 * (p[i-1] ^ (p[i-1] >> 30)) + i); 
		/* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
		/* In the previous versions, MSBs of the seed affect   */
		/* only MSBs of the array state[]. */
		/* 2002/01/09 modified by Makoto Matsumoto */
	}
	myMT_Next(mt);
}
#endif



DKC_MT_RAND32 *WINAPI dkcAllocMTRand32(uint32 seed)
{
	const uint32 alignMask = 15;	/* = 2^dkcdMT_RAND32_N - 1 */
	uint8 *tp = dkcAllocateFast_INL(sizeof(DKC_MT_RAND32) + alignMask);
	DKC_MT_RAND32 *p;
	int unit;
	if(NULL==tp) return NULL;
	//mymt.c16byte align technic for SSE
	p = (DKC_MT_RAND32 *)(((int)tp + alignMask) & ~alignMask);
	p->mallocPtr = (char *)tp;
#if 1
	unit = haveX86_UNIT();
	if (unit & MU_tSSE2) {
		p->function_ptr = (void *)dkcMTRand32Next_SSE2;
	} else if (unit & MU_tMMX) {
		p->function_ptr  = (void *)dkcMTRand32Next_MMX;
	} else {
		p->function_ptr  = (void *)dkcMTRand32Next_C;
	}
#else
	p->function_ptr  = (void *)dkcMTRand32Next_C;
#endif
	dkcMTRand32Init(p,seed);
	return p;
}

int WINAPI dkcFreeMTRand32(DKC_MT_RAND32 **p)
{
	dkcFreeFast_INL((void **)(&((*p)->mallocPtr)));
	return edk_SUCCEEDED;
}



uint32 WINAPI dkcMTRand32Get(DKC_MT_RAND32 *p){

	return dkcMTRand32Get_INL(p);
}