
;	fast Mersenne Twister routine for Pentium4
;	2002/10/19
;	MITSUNARI Shigeo(herumi@nifty.com)

%include "nasm.cfg"

%define N 624
%define M 397
%define MATRIX_A 0x9908B0DF
%define TEMP1 0x9D2C5680
%define TEMP2 0xEFC60000


	segment_data

	align 32
Q_80000000	dd	0x80000000, 0x80000000, 0x80000000, 0x80000000
Q_7FFFFFFF	dd	0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF
Q_MATRIX_A	dd	MATRIX_A, MATRIX_A, MATRIX_A, MATRIX_A
Q_TEMP1		dd	TEMP1, TEMP1, TEMP1, TEMP1
Q_TEMP2		dd	TEMP2, TEMP2, TEMP2, TEMP2

; defined int mymt.h
%define state	0
%define output	state+N*4
%define index	output+N*4

	segment_text

; input   : esi = base address
; output  : none
; destroy : eax, ebx, edx
; %1 : offset
; %2 : adj (M*4 or (M-N)*4)
%macro twist 2
		mov			eax, [esi+%1]		; u
		mov			ebx, [esi+%1+4]		; v
		mov			edx, ebx
		and			eax, 0x80000000
		inc			edx					; edx = v + 1
		and			ebx, 0x7FFFFFFF
		and			edx, 1				; (v & 1) ? 0 : 1
		or			eax, ebx
		dec			edx					; (v & 1) ? -1 : 0
		shr			eax, 1
		and			edx, MATRIX_A		; (v & 1) ? MATRIX_A : 0
		xor			eax, edx
		xor			eax, [esi+%1+%2]
		mov			[esi+%1], eax
%endm

%if 0	; ref.
;		movq		mm7, [Q_MATRIX_A]
;		movq		mm6, [Q_7FFFFFFF]
%macro twistD 2
		movq		mm0, [esi+%1]			; u
		movq		mm1, [esi+%1+4]			; v
		movq		mm2, mm0
		pxor		mm2, mm1
		pslld		mm1, 31
		pand		mm2, mm6
		pxor		mm2, mm0
		psrad		mm1, 31					; (v & 1) ? -1 : 0
		psrld		mm2, 1
		pand		mm1, mm7
		pxor		mm2, mm1
		pxor		mm2, [esi+%1+%2]
		movq		[esi+%1], mm2
%endm
%endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;dFmyMT_Next_SSE2  dkcMTRand32Next_SSE2ɖOύX
proc	dkcMTRand32Next_SSE2
		push		ebx
		push		esi
		push		edi
		push		ebp
%assign _P (4*4)
		mov			ebp, [esp+_P+4]

		movaps		xm7, [Q_MATRIX_A]
		movaps		xm6, [Q_7FFFFFFF]

		;--------- 1st loop ---------
		; N - M = 227 = 8 * 28 + 3
		lea			esi, [ebp+state+(N-M-3)*4]
		mov			ecx, -(N-M-3)

.lp1:
		movaps		xm0, [esi+ecx*4]		; [u3:u2:u1:u0]
		movaps		xm1, [esi+ecx*4+16]		; u4
		movaps		xm4, [esi+ecx*4+32]		; u4
		movaps		xm3, xm1
		movaps		xm2, xm0
		movaps		xm5, xm3
		pslldq		xm1, 12
		pslldq		xm4, 12
		psrldq		xm2, 4
		psrldq		xm5, 4
		por			xm1, xm2				; [u4:u3:u2:u1]
		por			xm4, xm5				; [u4:u3:u2:u1]

		movaps		xm2, xm0
		movaps		xm5, xm3
		pxor		xm2, xm1
		pxor		xm5, xm4
		pslld		xm1, 31
		pslld		xm4, 31
		pand		xm2, xm6
		pand		xm5, xm6
		pxor		xm2, xm0
		pxor		xm5, xm3
		movaps		xm0, [esi+ecx*4+(M-1)*4]	; [3:2:1:0]
		psrad		xm4, 31
		psrad		xm1, 31
		psrld		xm2, 1
		psrld		xm5, 1
		pand		xm1, xm7
		psrldq		xm0, 4						; [-:3:2:1]
		pand		xm4, xm7
		pxor		xm2, xm1
		movaps		xm1, [esi+ecx*4+(M+3)*4]	; 4
		movaps		xm3, xm1
		pxor		xm5, xm4
		movaps		xm4, [esi+ecx*4+(M+7)*4]	; 4
		pslldq		xm1, 12
		psrldq		xm3, 4						; [-:3:2:1]
		pslldq		xm4, 12
		por			xm1, xm0					; [4:3:2:1]
		por			xm4, xm3					; [4:3:2:1]
		pxor		xm2, xm1
		pxor		xm5, xm4
		movaps		[esi+ecx*4], xm2
		movaps		[esi+ecx*4+16], xm5
		add			ecx, byte 8
		jnz			near .lp1

		; remain 3
		twist		0, M*4
		twist		4, M*4
		twist		8, M*4

		;--------- 2nd loop ---------

		; M - 1 = 396 = 1 + 8 * 49 + 3
		lea			esi, [ebp+state+(N-4)*4]

		; start 1 line
		twist		-(M-4)*4, (M-N)*4

		; main loop
		; esi : 16byte alignment

		mov			ecx, -(M-5)

%if 1	; unloop
.lp2:
		movaps		xm0, [esi+ecx*4]		; [u3:u2:u1:u0]
		movaps		xm1, [esi+ecx*4+16]		; u4
		movaps		xm4, [esi+ecx*4+32]		; u4
		movaps		xm3, xm1
		movaps		xm2, xm0
		movaps		xm5, xm3
		pslldq		xm1, 12
		pslldq		xm4, 12
		psrldq		xm2, 4
		psrldq		xm5, 4
		por			xm1, xm2				; [u4:u3:u2:u1]
		por			xm4, xm5				; [u4:u3:u2:u1]

		movaps		xm2, xm0
		movaps		xm5, xm3
		pxor		xm2, xm1
		pxor		xm5, xm4
		pslld		xm1, 31
		pslld		xm4, 31
		pand		xm2, xm6
		pand		xm5, xm6
		pxor		xm2, xm0
		pxor		xm5, xm3
		movaps		xm0, [esi+ecx*4+(M-N-1)*4]	; [3:2:1:0]
		psrad		xm4, 31
		psrad		xm1, 31
		psrld		xm2, 1
		psrld		xm5, 1
		pand		xm1, xm7
		psrldq		xm0, 4						; [-:3:2:1]
		pand		xm4, xm7
		pxor		xm2, xm1
		movaps		xm1, [esi+ecx*4+(M-N+3)*4]	; 4
		movaps		xm3, xm1
		pxor		xm5, xm4
		movaps		xm4, [esi+ecx*4+(M-N+7)*4]	; 4
		pslldq		xm1, 12
		psrldq		xm3, 4						; [-:3:2:1]
		pslldq		xm4, 12
		por			xm1, xm0					; [4:3:2:1]
		por			xm4, xm3					; [4:3:2:1]
		pxor		xm2, xm1
		pxor		xm5, xm4
		movaps		[esi+ecx*4], xm2
		movaps		[esi+ecx*4+16], xm5
		add			ecx, byte 8
		jnz			near .lp2
%else
.lp2:
		movaps		xm0, [esi+ecx*4]		; [u3:u2:u1:u0]
		movaps		xm1, [esi+ecx*4+16]		; u4
		movaps		xm2, xm0
		pslldq		xm1, 12
		psrldq		xm2, 4
		por			xm1, xm2				; [u4:u3:u2:u1]

		movaps		xm2, xm0
		pxor		xm2, xm1
		pslld		xm1, 31
		pand		xm2, xm6
		pxor		xm2, xm0
		movaps		xm0, [esi+ecx*4+(M-N-1)*4]	; [3:2:1:0]
		psrad		xm1, 31
		psrld		xm2, 1
		pand		xm1, xm7
		psrldq		xm0, 4						; [-:3:2:1]
		pxor		xm2, xm1
		movaps		xm1, [esi+ecx*4+(M-N+3)*4]	; 4
		pslldq		xm1, 12
		por			xm1, xm0					; [4:3:2:1]
		pxor		xm2, xm1
		movaps		[esi+ecx*4], xm2
		add			ecx, byte 4
		jnz			.lp2
%endif

		; remain 3
		lea			esi, [ebp+state+(N-4)*4]
		twist		0, (M-N)*4
		twist		4, (M-N)*4
		twist		8, (M-N)*4

		lea			esi, [ebp+state+(N-1)*4]

		; last line

		mov			eax, [esi]			; u
		mov			ebx, [esi-(N-1) * 4]; v
		mov			edx, ebx
		and			eax, 0x80000000
		inc			edx					; edx = v + 1
		and			ebx, 0x7FFFFFFF
		and			edx, 1				; (v & 1) ? 0 : 1
		or			eax, ebx
		dec			edx					; (v & 1) ? -1 : 0
		shr			eax, 1
		and			edx, MATRIX_A		; (v & 1) ? MATRIX_A : 0
		xor			eax, edx
		xor			eax, [esi+(M-N) * 4]
		mov			[esi], eax

;-----------------------------------------------------------------------
myMT_Tempering_SSE2:

		; reset index
		xor			eax, eax
		mov			[ebp+index], eax

		lea			eax, [ebp+state+N*4]
		lea			edx, [ebp+output+N*4]
		movaps		xm7, [Q_TEMP1]
		movaps		xm6, [Q_TEMP2]
		mov			ecx, -N
		; N = 624 = 12 * 52
.lp_Tempering:
		movaps		xm0, [eax+ecx*4]
		movaps		xm2, [eax+ecx*4+16]
		movaps		xm4, [eax+ecx*4+32]
		movaps		xm1, xm0
		movaps		xm3, xm2
		movaps		xm5, xm4
		psrld		xm0, 11
		psrld		xm2, 11
		psrld		xm4, 11
		pxor		xm0, xm1
		pxor		xm2, xm3
		pxor		xm4, xm5
		movaps		xm1, xm0
		movaps		xm3, xm2
		movaps		xm5, xm4
		pslld		xm0, 7
		pslld		xm2, 7
		pslld		xm4, 7
		pand		xm0, xm7
		pand		xm2, xm7
		pand		xm4, xm7
		pxor		xm0, xm1
		pxor		xm2, xm3
		pxor		xm4, xm5
		movaps		xm1, xm0
		movaps		xm3, xm2
		movaps		xm5, xm4
		pslld		xm0, 15
		pslld		xm2, 15
		pslld		xm4, 15
		pand		xm0, xm6
		pand		xm2, xm6
		pand		xm4, xm6
		pxor		xm0, xm1
		pxor		xm2, xm3
		pxor		xm4, xm5
		movaps		xm1, xm0
		movaps		xm3, xm2
		movaps		xm5, xm4
		psrld		xm0, 18
		psrld		xm2, 18
		psrld		xm4, 18
		pxor		xm0, xm1
		pxor		xm2, xm3
		pxor		xm4, xm5
		movaps		[edx+ecx*4], xm0
		movaps		[edx+ecx*4+16], xm2
		movaps		[edx+ecx*4+32], xm4
		add			ecx, byte 12
		jnz			near .lp_Tempering

		pop			ebp
		pop			edi
		pop			esi
		pop			ebx
;CLKEND
		ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;		edx  mask ̍
;		edx = v; shr edx, 1; sbb edx, edx 
;		edx = v; inc edx; and edx, 1; dec edx; ̂ق
;
;dFmyMT_Next_MMXdkcMTRand32Next_MMXɖOύX
proc	dkcMTRand32Next_MMX
		push		ebx
		push		esi
		push		edi
		push		ebp

%assign _P (4*4)
		mov			ebp, [esp+_P+4]			; mt
		lea			esi, [ebp+state+(N-M-1)*4]

		mov			ecx, -(N-M-1)
.lp1:
		mov			eax, [esi+ecx*4]		; u
		mov			ebx, [esi+ecx*4+4]		; v
		mov			edx, ebx
		mov			edi, ebx
		and			eax, 0x80000000
		shl			edx, 31
		and			ebx, 0x7FFFFFFF
		sar			edx, 31					; (v & 1) ? -1 : 0
		or			eax, ebx
		mov			ebx, [esi+ecx*4+8]		; v
		shr			eax, 1
		and			edx, MATRIX_A			; (v & 1) ? MATRIX_A : 0
		xor			eax, edx
		and			edi, 0x80000000
		mov			edx, ebx
		and			ebx, 0x7FFFFFFF
		xor			eax, [esi+ecx*4+M*4]
		shl			edx, 31
		mov			[esi+ecx*4], eax

		sar			edx, 31					; (v & 1) ? -1 : 0
		or			edi, ebx
		shr			edi, 1
		and			edx, MATRIX_A			; (v & 1) ? MATRIX_A : 0
		xor			edi, edx
		xor			edi, [esi+ecx*4+M*4+4]
		mov			[esi+ecx*4+4], edi
		add			ecx, byte 2
		jnz			.lp1

		; remain
		mov			eax, [esi]				; u
		mov			ebx, [esi+4]			; v
		mov			edx, ebx
		and			eax, 0x80000000
		shl			edx, 31
		and			ebx, 0x7FFFFFFF
		sar			edx, 31					; (v & 1) ? -1 : 0
		or			eax, ebx
		shr			eax, 1
		and			edx, MATRIX_A			; (v & 1) ? MATRIX_A : 0
		xor			eax, edx
		xor			eax, [esi+M*4]
		mov			[esi], eax

		lea			esi, [ebp+state+(N-1)*4]
		mov			ecx, -(M-1)
.lp2:
		mov			eax, [esi+ecx*4]		; u
		mov			ebx, [esi+ecx*4+4]		; v
		mov			edx, ebx
		mov			edi, ebx
		and			eax, 0x80000000
		shl			edx, 31
		and			ebx, 0x7FFFFFFF
		sar			edx, 31					; (v & 1) ? -1 : 0
		or			eax, ebx
		mov			ebx, [esi+ecx*4+8]		; v
		shr			eax, 1
		and			edx, MATRIX_A			; (v & 1) ? MATRIX_A : 0
		xor			eax, edx
		and			edi, 0x80000000
		mov			edx, ebx
		and			ebx, 0x7FFFFFFF
		xor			eax, [esi+ecx*4+(M-N)*4]
		shl			edx, 31
		mov			[esi+ecx*4], eax

		sar			edx, 31					; (v & 1) ? -1 : 0
		or			edi, ebx
		shr			edi, 1
		and			edx, MATRIX_A			; (v & 1) ? MATRIX_A : 0
		xor			edi, edx
		xor			edi, [esi+ecx*4+(M-N)*4+4]
		mov			[esi+ecx*4+4], edi
		add			ecx, byte 2
		jnz			.lp2

		; last line

		mov			eax, [ebp+state+(N-1)*4]	; u
		mov			ebx, [ebp+state]			; v
		mov			edx, ebx
		and			eax, 0x80000000
		shl			edx, 31
		and			ebx, 0x7FFFFFFF
		sar			edx, 31					; (v & 1) ? -1 : 0
		or			eax, ebx
		shr			eax, 1
		and			edx, MATRIX_A			; (v & 1) ? MATRIX_A : 0
		xor			eax, edx
		xor			eax, [ebp+state+(M-1)*4]
		mov			[ebp+state+(N-1)*4], eax

;-----------------------------------------------------------------------
myMT_Tempering_MMX:

		emms
		; reset index
		xor			eax, eax
		mov			[ebp+index], eax

		lea			eax, [ebp+state+N*4]
		lea			edx, [ebp+output+N*4]
		movq		mm7, [Q_TEMP1]
		movq		mm6, [Q_TEMP2]
		mov			ecx, -N
		; N = 624 = 6 * 102
.lp_Tempering:
		movq		mm0, [eax+ecx*4]
		movq		mm2, [eax+ecx*4+8]
		movq		mm4, [eax+ecx*4+16]
		movq		mm1, mm0
		movq		mm3, mm2
		movq		mm5, mm4
		psrld		mm0, 11
		psrld		mm2, 11
		psrld		mm4, 11
		pxor		mm0, mm1
		pxor		mm2, mm3
		pxor		mm4, mm5
		movq		mm1, mm0
		movq		mm3, mm2
		movq		mm5, mm4
		pslld		mm0, 7
		pslld		mm2, 7
		pslld		mm4, 7
		pand		mm0, mm7
		pand		mm2, mm7
		pand		mm4, mm7
		pxor		mm0, mm1
		pxor		mm2, mm3
		pxor		mm4, mm5
		movq		mm1, mm0
		movq		mm3, mm2
		movq		mm5, mm4
		pslld		mm0, 15
		pslld		mm2, 15
		pslld		mm4, 15
		pand		mm0, mm6
		pand		mm2, mm6
		pand		mm4, mm6
		pxor		mm0, mm1
		pxor		mm2, mm3
		pxor		mm4, mm5
		movq		mm1, mm0
		movq		mm3, mm2
		movq		mm5, mm4
		psrld		mm0, 18
		psrld		mm2, 18
		psrld		mm4, 18
		pxor		mm0, mm1
		pxor		mm2, mm3
		pxor		mm4, mm5
		movq		[edx+ecx*4], mm0
		movq		[edx+ecx*4+8], mm2
		movq		[edx+ecx*4+16], mm4
		add			ecx, byte 6
		jnz			near .lp_Tempering

		emms
		pop			ebp
		pop			edi
		pop			esi
		pop			ebx
;CLKEND
		ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; original source is http://www.emit.jp/mt/mt.html
%if 0
proc	myMT_Next_other
		push		ebx
		push		esi
		push		edi
		push		ebp
%assign _P (4*4)
		mov			ebp, [esp+_P+4]
		emms

		mov			eax, MATRIX_A
		movd		mm4, eax
		punpckldq	mm4, mm4
		mov			eax, 1
		movd		mm5, eax
		punpckldq	mm5, mm5
		movq		mm6, mm5
		psllq		mm6, 31
		movq		mm7, mm6
		psubd		mm7, mm5

		lea			esi, [ebp+state+((N-M)/2)*8]
		mov			edi, esi
		add			edi, M*4
		
		mov			ecx, -((N-M)/2)

		movq		mm0, [esi+ecx*8]
		pand		mm0, mm6
		movq		mm1, [esi+ecx*8+4]
		movq		mm2, [edi+ecx*8]
		pand		mm1, mm7
loop0:
		por			mm0, mm1
		movq		mm3, mm1
		psrld		mm0, 1
		pand		mm3, mm5
		pxor		mm2, mm0
		psubd		mm3, mm5
		movq		mm0, [esi+ecx*8+8]
		pandn		mm3, mm4
		movq		mm1, [esi+ecx*8+12]
		pxor		mm3, mm2
		movq		mm2, [edi+ecx*8+8]
		pand		mm0, mm6
		movq		[esi+ecx*8], mm3
		pand		mm1, mm7
		inc			ecx
		jnz			loop0

		mov			eax, [esi]
		mov			ebx, [esi+4]
		and			eax, 0x80000000;UPPER_MASK
		and			ebx, 0x7FFFFFFF;LOWER_MASK
		mov			edx, [edi]
		or			eax, ebx
		shr			eax, 1
		or			ebx, 0xfffffffe
		xor			eax, edx
		add			ebx, 1
		xor			eax, MATRIX_A
		and			ebx, MATRIX_A
		xor			ebx, eax
		mov			[esi], ebx

		lea			esi, [ebp+state]
		mov			eax, [esi] ; mt[N] = mt[0];
		add			esi, N*4
		mov			[esi], eax ; mt[N] = mt[0];

		mov			edi, esi
		sub			edi, (N-M)*4
		add			esi, 4
		add			edi, 4
		
		mov			ecx, -((M+1)/2)

		movq		mm0, [esi+ecx*8]
		pand		mm0, mm6
		movq		mm1, [esi+ecx*8+4]
		movq		mm2, [edi+ecx*8]
		pand		mm1, mm7
loop1:
		por			mm0, mm1
		movq		mm3, mm1
		psrld		mm0, 1
		pand		mm3, mm5
		pxor		mm2, mm0
		psubd		mm3, mm5
		movq		mm0, [esi+ecx*8+8]
		pandn		mm3, mm4
		movq		mm1, [esi+ecx*8+12]
		pxor		mm3, mm2
		movq		mm2, [edi+ecx*8+8]
		pand		mm0, mm6
		movq		[esi+ecx*8], mm3
		pand		mm1, mm7
		inc			ecx
		jnz			loop1
		emms
		jmp			myMT_Tempering_SSE2
%endif
	end
